#ifndef __RT_LIB_RT_FIELD_MAP_H__
#define __RT_LIB_RT_FIELD_MAP_H__

#include <ostream>
#include <fire/private/prefix.h>

namespace FIRE_NAMESPACE {

class RT_API RT_FieldMap
{
  public:

  static const int MAX_FIELD_NAME_SIZE  = 128;
  static const int MAX_FIELD_NAME_DESCR = 256;

  enum FieldType {
    RT_NullType    = 0,
    RT_CharType    = 1,
    RT_StringType  = 2,
    RT_IntegerType = 3,
    RT_PriceType   = 4,
    RT_DecimalType = 5,
    RT_Reserved    = 6,
    RT_Enumeration = 7,
    RT_DateType    = 8,
    RT_TimeType    = 9,
    RT_Integer64   = 10
  };

  // The factory methods allow the only way to create a RT_FieldMap object
  static RT_FieldMap *create(const char *filename, bool make_default=true);
  static RT_FieldMap *create(const char *map_data, size_t map_size,
                             bool make_default=true);
  static void         destroy(RT_FieldMap *map);

  // set_default will return the old default.  Note: get_default returns
  // a const pointer, thus allowing the default map to be thread-safe.
  static       RT_FieldMap *set_default(const RT_FieldMap *new_default_map);
  static const RT_FieldMap *get_default();

  rt_errno append(const char *filename);
  rt_errno append(const char *map_data, size_t map_size);

  // Will return default_if_not_found if the field is not found.  The default
  // case means that get_name(...) will always return a  valid null-terminated
  // pointer.  If the application wants NULL, just pass in NULL as the 2nd arg.
  const char *get_name(int fid, const char *default_if_not_found = "") const;

  // All get_id(...)'s will return 0 if fid not found
  int get_id(const char *name)                         const;
  int get_id(const char *name, FieldType *type)        const;

  // get_type() will return RT_NullType if fid is not found
  FieldType get_type(int fid) const;

  private:
    // Clients must use create(...)
    RT_FieldMap();
   ~RT_FieldMap();
    RT_FieldMap(const RT_FieldMap &not_copyable);
    RT_FieldMap &operator=(const RT_FieldMap &not_copyable);
    class RT_FieldMapImpl * const m_impl;

  public:
    unsigned int count() const;
    const char *get_name_by_idx(unsigned int idx) const;
    int         get_id_by_idx  (unsigned int idx) const;
    FieldType   get_type_by_idx(unsigned int idx) const; 
};

//
// Implementation
//
static inline std::ostream& operator<<(std::ostream &os, const RT_FieldMap &map)
{
  os << "Dump FieldMap num_fids=> " << map.count() << "\n";
  for (unsigned int i=0; i<map.count(); ++i)
    os << "  FieldName: " << map.get_name_by_idx(i) << 
          " Id:" << map.get_id_by_idx(i) << 
          " Type:" << map.get_type_by_idx(i) << "\n";
  return os;
}

} // namespace

#endif
