#include <string>
#include <memory>
#include <algorithm>
#include "FDS/rt_api.h"
#include "FDS/rt_fields.h"
#include "FDS/Level2Feature.h"

using namespace std;
using namespace FDS;

class RTClient
{
  public:

  RTClient(const string &connection);
  RTClient(const string& user, const string& host, const string& path);
  int start();
  int stop();
  int make_request(const string &symbol);
  int main_loop();

  static void handle_message(int tag, const RT_Message *msg,
                              const RT_Record   *rec, void *closure);

  static void handle_connection_change(bool is_connected, const RT_Message *msg,
                                       void *closure);
  // our state is pretty simple
  string connection_string;
  bool   exit_main_loop;
  int    num_messages;
  // OTP variables
  string user_serial;
  string otp_path;
  string host;
  bool is_otp;

  private:
  RTClient(const RTClient &not_copyable);
  RTClient& operator=(const RTClient &not_copyable);
};

RTClient::RTClient(const string &connection) : connection_string(connection)
{
  is_otp = false;
  exit_main_loop = false;
  num_messages = 0;
  FDF::set_connection_ctrl_cb(&RTClient::handle_connection_change, this);
}

RTClient::RTClient(const string& user, const string& host_name, const string& path) : user_serial(user), otp_path(path) , host(host_name) {
  is_otp = true;
  exit_main_loop = false;
  num_messages = 0;
  FDF::set_connection_ctrl_cb(&RTClient::handle_connection_change, this);
}
//
// Look up installation directory for Windows Clients
//
#define CONFIGURATION_URI "reg:/HKEY_LOCAL_MACHINE|HKEY_CURRENT_USER/Software/FactSet/FDF"

int RTClient::start()
{
  FDF::load_properties(CONFIGURATION_URI);
  
  string log_file    = "text_quote.log";
  string fields_file = "../../../etc/rt_fields.xml";

  // Get our base installation directory from the properties (loaded from above)
  char base_dir[FDF::MAX_PROPERTY_SIZE];
  if (FDF::get_property("RT_INSTALLATION_DIR", base_dir)) {
    log_file = string(base_dir) + "/log/" + log_file;
    fields_file = string(base_dir) + "/etc/rt_fields.xml";
  }

  // Uncomment to open a log file, by default this program writes to the console
  // FDF::log_open(log_file.c_str());

  // Open our map file
  RT_FieldMap *map = RT_FieldMap::create(fields_file.c_str());
  if (!map)
    RT_LOG_WARN("Unable to open the field map file: " << fields_file);
  rt_errno err;
  if (!is_otp) {
    // Set the connection info (user:pass@host:port), from the connection string
    err = FDF::set_connection_info(connection_string.c_str());
    if (err) {
      RT_LOG_ERROR("Unable to get connection info from registry e:" << err);
      return err;
    }
  }
  else {
    size_t sep_loc = otp_path.find_last_of("/\\");
    string path = "";
    string device_id = "";

    if (sep_loc != std::string::npos)
    {
      path = otp_path.substr(0, sep_loc);
      device_id = otp_path.substr(sep_loc + 1);
    }
    else
    {
      device_id = otp_path;
    }

    // Removing .data extension from the key-counter data file
    size_t ext_loc = device_id.find(".data\0");
    if (ext_loc != std::string::npos)
    {
      device_id = device_id.substr(0, ext_loc);
    }

    err = FDF::set_connection_info(host.c_str(),
      user_serial.c_str(),
      device_id.c_str(),
      NULL,
      NULL,
      path.c_str(),
      false);
    if (err) {
      RT_LOG_ERROR("Unable to connect using OTP e:" << err);
      return err;
    }
  }
    
  // Connect to the data server
  err = FDF::connect();
  if (err) {
    RT_LOG_ERROR("Connection to RT_Server[" << FDF::get_connection_info()
                 << "] failed err:" << err);
    return err;
  }
  return 0;
}

int RTClient::stop()
{
  return FDF::disconnect();
}

// Static method, just log the message
void RTClient::handle_connection_change(bool is_connected, const RT_Message *msg,
                                      void *closure)
{
  RT_LOG_INFO("Control Message: " << *msg);
  if (!is_connected) RT_LOG_ERROR("server disconnected!! :(");
}

// Static method, just print the message to our log file
void RTClient::handle_message(int tag, const RT_Message *msg, 
                                        const RT_Record   *rec, void *closure)
{
  RTClient *clnt = static_cast<RTClient *>(closure);

  ++(clnt->num_messages);
  RT_LOG_INFO("message#" << clnt->num_messages << " tag:" << tag <<
              " msg: " << *msg);

  // Uncomment to print the current cached record as well:
  // RT_LOG_INFO("record--->>>>" << *rec);

  if (msg->is_closed()) {
    RT_LOG_ERROR("Server closed stream");
    FDF::cancel(tag);
    clnt->exit_main_loop = true;
    if (msg->is_error()) RT_LOG_ERROR(msg->get_error_description(NULL));
  }

  if (msg->is_stale())
    RT_LOG_INFO("STALE!!!!! reason: " << msg->get_stale_reason(NULL));
}

// Make a real-time request
int RTClient::make_request(const string &symbol)
{
  int tag;
  // A blank service will extract it from the symbol (if there is one)
  RT_Request req("", symbol.c_str());
  int err = FDF::request(req, &RTClient::handle_message, this, &tag);

  if (err)
    RT_LOG_WARN(req << " err:" << err);
  else
    RT_LOG_INFO(req << " tag:" << tag);
  return err;
}

int RTClient::main_loop()
{
  int err = 0;
  while (!exit_main_loop) {
    err = FDF::dispatch(-1);
    if (err) {
      RT_LOG_ERROR("Error in dispatch e:" << err);
      break;
    }
  }
  return err;
}

// Helper class for parsing the command line arguments
class InputParser {
public:
  InputParser(int &argc, char **argv) {
    for (int i = 1; i < argc; ++i)
      this->tokens.push_back(string(argv[i]));
  }
  vector<string> get_multiple_items(const string &option) const {
    vector<string>::const_iterator itr;
    vector<string> symbols;
    itr = find(this->tokens.begin(), this->tokens.end(), option);
    ++itr;
    while (itr != this->tokens.end()) {
      if (itr->at(0) == '-') break; // dont add if it is a command
      symbols.push_back(*itr);
      ++itr;
    }
    return symbols;
  }

  const string get_single_item(const string &option) const {
    vector<string>::const_iterator itr;
    itr = find(this->tokens.begin(), this->tokens.end(), option);
    if (itr != this->tokens.end() && ++itr != this->tokens.end()) {
      return *itr;
    }
    return string("");
  }

  bool is_option_exists(const string &option) const {
    return find(this->tokens.begin(), this->tokens.end(), option)
      != this->tokens.end();
  }
private:
  vector <string> tokens;
};

//
// example usage:
// .\TextQuote -c user:passwd@api-stage.df.factset.com -s FDS1::IBM-USA
// .\TextQuote -u user-serial -p ./AAAA.data -h api-stage.df.factset.com -s FDS1::IBM-USA
//
int main(int argc, char **argv)
{
  if (argc <= 4) {
    RT_LOG_RAW("Basic authentication Usage: " << argv[0] << "-c {user-serial:passwd@host} -s {symbol} {symbol} .. {symbol}\n");
    RT_LOG_RAW("Example: TextQuote -c user:passwd@api-stage.df.factset.com -s FDS1::IBM-USA\n");
    RT_LOG_RAW("OTP authentication \n Usage: " << argv[0] << "\n -u {user-serial} \n -p {full-path-to-OTP-file} \n -h {host} \n -s {symbol} {symbol} .. {symbol}\n");
    RT_LOG_RAW("Example: TextQuote -u user-serial -p ./AAAA.data -h api-stage.df.factset.com -s FDS1::IBM-USA\n");
    RT_LOG_RAW("-l2 option is used to enable level 2 data\n");
    return 3;
  }

  InputParser input(argc, argv);
  auto_ptr<RTClient> clnt;
  if (input.is_option_exists("-p")) { // otp authentication
    string user_serial = input.get_single_item("-u");
    string otp_path = input.get_single_item("-p");
    string host = input.get_single_item("-h");
    if (user_serial.empty() || otp_path.empty() || host.empty()) {
      RT_LOG_RAW("For OTP please provide all three options ( -p -u -h )\n");
      return 3;
    }
    clnt = auto_ptr<RTClient>(new RTClient(user_serial, host, otp_path));
  }
  else { // basic authentication
    string connection_string = input.get_single_item("-c");
    if (connection_string.empty()) {
      RT_LOG_RAW("Connection string is empty. Please add a valid connection string with option -c\n");
      return 3;
    }
    clnt = auto_ptr<RTClient>(new RTClient(connection_string));
  }

  RT_LOG_INFO("INFO: about to connect");
  int err = clnt->start();
  if (err) {
    RT_LOG_ERROR("Error in connect e: " << err);
    return err;
  }
  RT_LOG_INFO("connected, make request");

  vector<string> symbols = input.get_multiple_items("-s");
  bool is_level2 = input.is_option_exists("-l2");
  if(is_level2) {
    if (!FDF::is_active(Level2Feature())) {
      bool res = FDF::enable(Level2Feature());
      if (!res) {
        RT_LOG_ERROR("Error to enable Level 2 feature.");
        is_level2 = false;
      }
    }
  }
  while (!symbols.empty()) {
    string cur_symbol = symbols.back();
    if(is_level2 && cur_symbol.find(":L2") == string::npos)
      cur_symbol += ":L2"; // add only if ":L2" is not already present
    clnt->make_request(cur_symbol);
    symbols.pop_back();
  }
  // Enter main loop
  int rtn = clnt->main_loop();

  // disconnect
  RT_LOG_INFO("main loop rtn'd: " << rtn);
  clnt->stop();
}

