#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <memory>
#include <algorithm>
#include "FDS/rt_api.h"
#include "FDS/rt_fields.h"

using namespace std;
using namespace FDS;

class RTClient
{
  public:

  RTClient(const string &connection);
  RTClient(const string& user, const string& host_name, const string& path);
  int start();
  int stop();
  int make_request(const string &symbol);
  int main_loop();

  static void handle_response(int tag, const RT_Message *msg,
                              const RT_Record *rec, void *closure);
  static void handle_connection_change(bool is_connected, const RT_Message *msg,
                                       void *closure);
  string connection_string;
  bool   exit_main_loop;
  int    num_reqs, num_msgs;
  ostream    *ostrm;
  vector<int> fids;
  vector<int> tags;
  // OTP variables
  string user_serial;
  string otp_path;
  string host;
  bool is_otp;
};

RTClient::RTClient(const string &connection) : connection_string(connection),
  exit_main_loop(false), num_reqs(0), num_msgs(0), is_otp(false)
{
  FDF::set_connection_ctrl_cb(&RTClient::handle_connection_change, this);

  // If you want to output to a file, just open the file and set it here
  // Example: ostrm = new ofstream("recording.csv");
  ostrm = &cout;
}
RTClient::RTClient(const string& user, const string& host_name, const string& path) : user_serial(user), otp_path(path) , host(host_name),
exit_main_loop(false), num_reqs(0), num_msgs(0), is_otp(true)
{
  FDF::set_connection_ctrl_cb(&RTClient::handle_connection_change, this);
  // If you want to output to a file, just open the file and set it here
  // Example: ostrm = new ofstream("recording.csv");
  ostrm = &cout;
}
//
// Look up installation directory for Windows Clients
//
#define CONFIGURATION_URI "reg:/HKEY_LOCAL_MACHINE|HKEY_CURRENT_USER/Software/FactSet/FDF"

int RTClient::start()
{
  FDF::load_properties(CONFIGURATION_URI);

  string log_file = "tick_analyzer.log";
  string fields_file = "../../../etc/rt_fields.xml";

  FDF::log_open(log_file.c_str());

  // Open our field map file
  RT_FieldMap *fmap = RT_FieldMap::create(fields_file.c_str());
  if (!fmap) {
    RT_LOG_WARN("Unable to open field map file: " << fields_file);
    return 1;
  }

  rt_errno err;
  if (!is_otp) {
    // Set the connection info (user:pass@host:port), from the connection string
    err = FDF::set_connection_info(connection_string.c_str());
    if (err) {
      RT_LOG_ERROR("Unable to get connection info from registry e:" << err);
      return err;
    }
  }
  else {
    size_t sep_loc = otp_path.find_last_of("/\\");
    string path = "";
    string device_id = "";

    if (sep_loc != std::string::npos)
    {
      path = otp_path.substr(0, sep_loc);device_id = otp_path.substr(sep_loc + 1);
    }
    else
    {
      device_id = otp_path;
    }

    // Removing .data extension from the key-counter data file
    size_t ext_loc = device_id.find(".data\0");
    if (ext_loc != std::string::npos)
    {
      device_id = device_id.substr(0, ext_loc);
    }

    err = FDF::set_connection_info(host.c_str(),
      user_serial.c_str(),
      device_id.c_str(),
      NULL,
      NULL,
      path.c_str(),
      false);
    if (err) {
      RT_LOG_ERROR("Unable to connect using OTP e:" << err);
      return err;
    }
  }
    
  // Connect to the data server
  err = FDF::connect();
  if (err) {
    RT_LOG_ERROR("Connection to RT_Server[" << FDF::get_connection_info()
                 << "] failed err:" << err);
    return err;
  }
  return 0;
}

int RTClient::stop()
{
  // Cancel all outstanding tags
  vector<int>::iterator it   = tags.begin();
  vector<int>::iterator it_e = tags.end();
  for ( ; it != it_e ; ++it)
    FDF::cancel(*it);
  tags.clear();

  fids.clear();
  exit_main_loop = false;
  num_reqs = 0;
  num_msgs = 0;

  return FDF::disconnect();
}

void RTClient::handle_connection_change(bool is_connected, const RT_Message *msg,
                                      void *closure)
{
  RT_LOG_INFO("Control Message: " << *msg);
  if (!is_connected) RT_LOG_ERROR("server disconnected");
}

void RTClient::handle_response(int tag, const RT_Message *msg, 
                                        const RT_Record   *rec, void *closure)
{
  RTClient *clnt = static_cast<RTClient *>(closure);

  // Check if this is an initial response 
  // (snapshot of all fields in their most up-to-date state)
  if (msg->is_response()) {

    // Make a list of all unique fields
    for (unsigned int i=0 ; i<(unsigned int)msg->count() ; ++i) {
      int fid = msg->get_by_idx(i).fid;
      vector<int>::iterator it = find(clnt->fids.begin(), clnt->fids.end(), fid);
      if (it == clnt->fids.end())
        clnt->fids.push_back(fid);
    }

    ++clnt->num_msgs;

    // Check if this is the final initial response
    if (clnt->num_msgs == clnt->num_reqs) {

      // Write the header line
      *(clnt->ostrm) << "KEY";
      if (clnt->fids.size())
        *(clnt->ostrm) << ',';

      for (unsigned int i=0 ; i<clnt->fids.size() ; ++i) {
        *(clnt->ostrm) << RT_FieldMap::get_default()->get_name(clnt->fids[i]);
        if (i < clnt->fids.size()-1)
          *(clnt->ostrm) << ',';
        else
          *(clnt->ostrm) << endl;
      }

      // Write out all the initial responses (& set MSG_TYPE to #INITIAL#)
      for (unsigned int i=0 ; i<clnt->tags.size() ; ++i) {
        const RT_Record *rec;
        FDF::get_record(clnt->tags[i], &rec, NULL);

        *(clnt->ostrm) << rec->get_key().to_string();
        if (clnt->fids.size())
          *(clnt->ostrm) << ',';
        
        for (unsigned int j=0; j<clnt->fids.size() ; ++j) {
          if (clnt->fids[j] == FIDS::MSG_TYPE) {
            *(clnt->ostrm) << "#INITIAL#,";
            continue;
          }
          *(clnt->ostrm) << rec->get(clnt->fids[j]).to_string();
          if (j < clnt->fids.size()-1)
            *(clnt->ostrm) << ',';
          else
            *(clnt->ostrm) << endl;
        }
        
      }
    }
  }
  else if (clnt->num_msgs == clnt->num_reqs) {
    // Once all initial responses have been handled, print out updates
    *(clnt->ostrm) << msg->get_key();
    if (clnt->fids.size())
      *(clnt->ostrm) << ',';

    for (unsigned int i=0 ; i<clnt->fids.size() ; ++i) {
      *(clnt->ostrm) << msg->get(clnt->fids[i]).to_string();
      if (i < clnt->fids.size()-1)
        *(clnt->ostrm) << ',';
      else
        *(clnt->ostrm) << endl;
    }
  }
  clnt->ostrm->flush();

  if (msg->is_closed()) {
    RT_LOG_ERROR("Server closed stream; tag: " << tag 
                 << ", key: " << msg->get_key());
    if (msg->is_error()) RT_LOG_ERROR(msg->get_error_description(NULL));
    FDF::cancel(tag);
    // Remove tag from tags list
    vector<int>::iterator it = find(clnt->tags.begin(), clnt->tags.end(), tag);
    if (it != clnt->tags.end())
      clnt->tags.erase(it);
  }
  if (msg->is_stale())
    RT_LOG_INFO("STALE! reason: " << msg->get_stale_reason(NULL));
}

int RTClient::make_request(const string &symbol)
{
  int tag;
  RT_Request req("", symbol.c_str());
  int err = FDF::request(req, &RTClient::handle_response, this, &tag);
  if (err) {
    RT_LOG_WARN(req << " err:" << err);
  }
  else {
    ++num_reqs;
    tags.push_back(tag);
    RT_LOG_INFO(req << " tag:" << tag);
  }
  return err;
}

int RTClient::main_loop()
{
  int err = 0;
  while (!exit_main_loop) {
    err = FDF::dispatch(-1);
    if (err) {
      RT_LOG_ERROR("Error in dispatch e:" << err);
      break;
    }
  }
  return err;
}

// Helper class for parsing the command line arguments
class InputParser {
public:
  InputParser(int &argc, char **argv) {
    for (int i = 1; i < argc; ++i)
      this->tokens.push_back(string(argv[i]));
  }
  vector<string> get_multiple_items(const string &option) const {
    vector<string>::const_iterator itr;
    vector<string> symbols;
    itr = find(this->tokens.begin(), this->tokens.end(), option);
    ++itr;
    while (itr != this->tokens.end()) {
      if (itr->at(0) == '-') break; // dont add if it is a command
      symbols.push_back(*itr);
      ++itr;
    }
    return symbols;
  }

  const string get_single_item(const string &option) const {
    vector<string>::const_iterator itr;
    itr = find(this->tokens.begin(), this->tokens.end(), option);
    if (itr != this->tokens.end() && ++itr != this->tokens.end()) {
      return *itr;
    }
    return string("");
  }

  bool is_option_exists(const string &option) const {
    return find(this->tokens.begin(), this->tokens.end(), option)
      != this->tokens.end();
  }
private:
  vector <string> tokens;
};


//
// example usage:
// .\tick_analyzer -c user:passwd@api-stage.df.factset.com -s FDS1::IBM-USA FDS1::GOOG-USA
// .\tick_analyzer -u user-serial -p .\AAAA.data -h api-stage.df.factset.com -s FDS1::IBM-USA
//
int main(int argc, char **argv)
{
  if (argc <= 4) {
    RT_LOG_RAW("Basic authentication Usage: " << argv[0] << "-c {user-serial:passwd@host} -s {symbol} {symbol} .. {symbol}\n");
    RT_LOG_RAW("Example: tick_analyzer -c user:passwd@api-stage.df.factset.com -s FDS1::IBM-USA\n");
    RT_LOG_RAW("OTP authentication \n Usage: " << argv[0] << "\n -u {user-serial} \n -p {full-path-to-OTP-file} \n -h {host} \n -s {symbol} {symbol} .. {symbol}\n");
    RT_LOG_RAW("Example: tick_analyzer -u user-serial -p ./AAAA.data -h api-stage.df.factset.com -s FDS1::IBM-USA\n");
    return 3;
  }
  
  InputParser input(argc, argv);
  auto_ptr<RTClient> clnt;
  if (input.is_option_exists("-p")) { // otp authentication
    string user_serial = input.get_single_item("-u");
    string otp_path = input.get_single_item("-p");
    string host = input.get_single_item("-h");
    if (user_serial.empty() || otp_path.empty() || host.empty()) {
      RT_LOG_RAW("For OTP please provide all three options ( -p -u -h )\n");
      return 3;
    }
    clnt = auto_ptr<RTClient>(new RTClient(user_serial, host, otp_path));
  }
  else { // basic authentication
    string connection_string = input.get_single_item("-c");
    if (connection_string.empty()) {
      RT_LOG_RAW("Connection string is empty. Please add a valid connection string with option -c\n");
      return 3;
    }
    clnt = auto_ptr<RTClient>(new RTClient(connection_string));
  }
  int err = clnt->start();
  if (err) {
    RT_LOG_ERROR("Error in connect e: " << err);
    return err;
  }
  RT_LOG_INFO("connected, make request");

  vector<string> symbols = input.get_multiple_items("-s");
  while (!symbols.empty()) {
    clnt->make_request(symbols.back());
    symbols.pop_back();
  }

  // Enter main loop
  int rtn = clnt->main_loop();

  // disconnect
  RT_LOG_INFO("main loop rtn'd: " << rtn);
  clnt->stop();
}
